"""
pytest-plugin API
=================

The plugin provides a cijoe-instance readily available as a test-fixture, setup per
test with a nodeid-defined output-directory. For example::

    def test_foo(cijoe):
        err, _ = cijoe.run("hostname")
        asssert not err

To provide the cijoe-instance a configuration and output directory must be provided.
These are given via pytest, e.g.::

    pytest --config default.toml --output /tmp/foo

In case no arguments are provided, defaults are used.
"""

from pathlib import Path

import pytest

from cijoe.core.command import Cijoe, default_output_path
from cijoe.core.resources import Collector, Config

pytest.cijoe_instance = None


def pytest_addoption(parser):
    """
    Add options ``--config`` and ``--output`` to pytest, these will be used for the
    instantiation of cijoe.
    """

    collector = Collector()
    collector.collect()

    parser.addoption(
        "--config",
        action="store",
        type=Path,
        help="Path to cijoe configuration",
        default=str(collector.resources["configs"]["core.example_config_default"]),
    )
    parser.addoption(
        "--output",
        action="store",
        type=Path,
        help="Path to cijoe output directory",
        default=default_output_path(),
    )


def pytest_configure(config):
    """
    Initializes the cijoe instance using pytest-options ``--config`` and ``--output``

    The cijoe-instance is stored in ``pytest.cijoe_instance``, this might appear as bad
    form. However, it is a common pytest-pattern for enabling access to state otherwise
    only accessible to tests and fixtures.

    Why would this be needed? Well, for large paramaterizations, e.g. to generate input
    to pytest.mark.parametrize(). Here it is convenient to be able to access the same
    cijoe instance, and for example generate test-parametrization based on the content
    of the cijoe-configuration.
    """

    cijoe_config_path = config.getoption("--config")

    cijoe_config = Config.from_path(cijoe_config_path)
    if cijoe_config is None:
        raise Exception(f"Failed loading config({cijoe_config_path})")

    pytest.cijoe_instance = Cijoe(
        cijoe_config,
        config.getoption("--output"),
        False,
    )
    if pytest.cijoe_instance is None:
        raise Exception("Failed instantiating cijoe")


def pytest_terminal_summary(terminalreporter, exitstatus, config):
    """
    Prints out a notice that cijoe is beeing used along with the values of the
    ``--config`` and ``--output`` options.
    """

    terminalreporter.ensure_newline()
    terminalreporter.section(
        "-={[ CIJOE pytest-plugin ]}=-", sep="-", blue=True, bold=True
    )
    terminalreporter.line("config: %r" % config.getoption("--config"))
    terminalreporter.line("output: %r" % config.getoption("--output"))


@pytest.fixture
def cijoe(request):
    """
    Provides a cijoe-instance, initialized with the pytest-options: ``--config``, and
    ``--output`` and with a per-test customization of the output directory.
    """

    if pytest.cijoe_instance is None:
        raise Exception("Invalid configuration or instance")

    pytest.cijoe_instance.set_output_ident(request.node.nodeid)

    return pytest.cijoe_instance



"""
insert / remove null_blk
========================

Insert or remove null_blk instances, based on the value of args.do

Retargetable: True
------------------
"""

import errno
from argparse import ArgumentParser

import cijoe.linux.null_blk as null_blk


def add_args(parser: ArgumentParser):
    parser.add_argument(
        "--do",
        choices=["insert", "remove"],
        default="insert",
        help="The commands to be run on the nullblk module.",
    )


def main(args, cijoe):
    """Insert or remove the null_blk"""

    do = args.do
    if do == "insert":
        err, _ = null_blk.insert(cijoe)
    elif do == "remove":
        err, _ = null_blk.remove(cijoe)
    else:
        err = errno.EINVAL

    return err



#!/usr/bin/env python3
"""
Wait for qemu guest termination
===============================

Note: The script will not fail if the guest does not exist.
Note: This script does not itself terminate the qemu guest.

Retargetable: False
-------------------
"""
import logging as log
import time
from argparse import ArgumentParser

from cijoe.qemu.wrapper import Guest


def add_args(parser: ArgumentParser):
    parser.add_argument("--guest_name", type=str, help="Name of the qemu guest.")
    parser.add_argument(
        "--timeout",
        type=int,
        default=60,
        help="Amount of seconds to wait for the qemu guest to terminate.",
    )


def main(args, cijoe):
    """Wait for termination of qemu guest"""

    if "guest_name" not in args:
        log.error("missing argument: guest_name")
        return 1

    guest = Guest(cijoe, cijoe.config, args.guest_name)

    start = time.time()
    err, terminated = guest.wait_for_termination(args.timeout)
    end = time.time()

    if terminated:
        log.info(
            f"Guest({args.guest_name}) terminated gracefully in {end-start} seconds"
        )
    else:
        log.warning(
            f"Guest({args.guest_name}) was not terminated within {args.timeout} seconds"
        )

    return err



import argparse
import errno
import logging as log
import os
import re
import shutil
import sys
import tempfile
import time
from pathlib import Path
from typing import Any, Optional

import jinja2

import cijoe.core
from cijoe.core.command import Cijoe, default_output_path
from cijoe.core.resources import (
    Config,
    Workflow,
    dict_from_tomlfile,
    dict_from_yamlfile,
    dict_substitute,
    dict_to_tomlfile,
    get_resources,
)

DEFAULT_CONFIG_FILENAME = "cijoe-config.toml"
DEFAULT_WORKFLOW_FILENAME = "cijoe-workflow.yaml"
DEFAULT_SCRIPT_FILENAME = "cijoe-script.py"
SEARCH_PATHS = [
    Path.cwd(),
    Path.cwd() / ".cijoe",
    Path.home() / ".config" / "cijoe",
    Path.home() / ".cijoe",
]
STEP_NAME_REGEX = r"^(?!\d)[a-z0-9_-]+$"


class OverrideDefaultAction(argparse._StoreAction):
    def __call__(self, parser, namespace, values, option_string=None):
        self.default = values
        setattr(namespace, self.dest, values)


class AppendDefaultAction(argparse._AppendAction):
    def __call__(self, parser, namespace, values, option_string=None):
        self.default.append(values)
        super().__call__(parser, namespace, values, option_string)


def search_for_file(path: Path) -> Optional[Path]:
    """
    Search for a file named path.name. In case the given 'path' does not
    have a directory path, then SEARCH_PATH is used.

    Returns a resolved Path object on success and None on error.
    """

    # Has a directory part; resolve it and check that it exists
    if path.is_absolute() or len(path.parts) != 1:
        path = path.resolve()
        if path.exists():
            return path

        return None

    # Only a filename; look in search paths
    for spath in SEARCH_PATHS:
        candidate = (spath / path).resolve()
        if candidate.exists():
            return candidate

    return None


def create_combined_toml(configs: list[Path], path: Path):
    combined_config: dict[str, Any] = {}
    for config in configs:
        combined_config |= dict_from_tomlfile(config)

    dict_to_tomlfile(combined_config, path)


def log_errors(errors):
    for error in errors:
        log.error(error)


def cli_integrity_check(args):
    """Lint a workflow"""

    log.info("cli: lint")
    log.info(f"workflow: '{args.workflow}'")
    log.info(f"config: '{args.config}'")

    workflow_dict = dict_from_yamlfile(args.workflow.resolve())

    steps = workflow_dict.get("steps", None)
    if steps is None:
        log.error("No steps in workflow, nothing to do.")
        return errno.EINVAL

    for nr, step in enumerate(steps, 1):
        step_name = step.get("name", None)
        if step_name is None:
            log.error(f"Step number({nr}) is missing mandatory 'name'")
            return errno.EINVAL

        if not re.match(STEP_NAME_REGEX, step_name):
            log.error(
                f"Invalid step_name({step_name}); must match regex({STEP_NAME_REGEX})"
            )
            return errno.EINVAL

    errors = Workflow.dict_normalize(workflow_dict)  # Normalize it
    errors += Workflow.dict_lint(args, workflow_dict)  # Check the yaml-file

    with tempfile.NamedTemporaryFile() as config_file:
        create_combined_toml(args.config, Path(config_file.name))
        args.config = Path(config_file.name)

        config = Config.from_path(args.config)
        if not config:
            log.error(f"failed: Config.from_path({args.config})")
            return errno.EINVAL

        errors += dict_substitute(workflow_dict, config.options)

    if errors:
        log_errors(errors)
        log.error("failed: 'see errors above'; Failed")
        return errno.EINVAL

    return 0


def cli_resources(args):
    """List the reference configuration files provided with cijoe packages"""

    log.info("cli: resources")
    resources = get_resources()

    print("Resources collected by the CIJOE collector are listed below.")
    for category, category_resources in sorted(resources.items()):
        print(f"{category}:" + ("" if category_resources.items() else " ~"))

        for ident, path in sorted(category_resources.items()):
            print(f"  - ident: {ident}")
            print(f"    path: {path}")

    return 0


def cli_archive(args):
    """Move 'output' directory into archive"""

    if args.output.exists():
        state_path = args.output / "workflow.state"
        tag = ""
        if state_path.exists():
            state = dict_from_yamlfile(state_path)
            tag = "" if state.get("tag", None) is None else "-".join(state["tag"]) + "-"
        t = str(time.strftime("%Y-%m-%d_%H:%M:%S"))
        archive = args.output.with_name("cijoe-archive") / f"{tag}{t}"
        os.makedirs(archive)
        log.info(f"moving existing output-directory({args.output}) to '{archive}'")
        os.rename(args.output, archive)


def cli_produce_report(args):
    """Produce workflow-report"""

    config_path = args.output / "config.orig"
    state_path = args.output / "workflow.state"

    if not state_path.exists():
        log.error("no workflow.state, nothing to produce a report for")
        return errno.EINVAL

    if not config_path.exists():
        log.error("missing config")
        return errno.EINVAL

    # Check config/substitutions
    config = Config.from_path(config_path)
    if not config:
        log.error(f"failed: Config.from_path({config_path})")
        return errno.EINVAL

    cijoe = Cijoe(config, args.output, args.monitor)

    resources = get_resources()

    reporter = resources["scripts"]["core.reporter"]
    if reporter.func is None:
        reporter.load()

    setattr(args, "report_open", args.skip_report)

    return reporter.func(args, cijoe)


def cli_example(args):
    """Create example config.toml and workflow.yaml"""
    log.info("cli: examples")

    resources = get_resources()

    # Print examples when called like "cijoe --example"
    if args.example == "list_examples_and_exit":
        for workflow_name, workflow in sorted(resources.get("workflows").items()):
            if "example" not in workflow_name:
                continue

            pkg_name, tail = workflow_name.split(".", 1)
            example_name = tail.replace("example_workflow_", "")

            print(f"{pkg_name}.{example_name}")
        return 0

    pkg_name, *tail = args.example.split(".")
    if len(tail) > 1:
        log.error(f"Invalid argument: {args.example}")
        return errno.EINVAL

    # Emit examples
    for workflow_name, workflow in sorted(resources.get("workflows").items()):
        if not workflow_name.startswith(pkg_name):  # Not the requested package
            continue
        if "example" not in workflow_name:  # Not an example workflow
            continue

        cur_example_id = workflow_name.replace("example_workflow_", "")
        cur_pkg_name, cur_example_name = cur_example_id.split(".", 1)
        cur_example_dir = Path.cwd() / f"cijoe-example-{cur_example_id}"

        log.info(f"cur_pkg_name{cur_pkg_name} cur_example_name{cur_example_name}")
        if tail and "".join(tail) != cur_example_name:
            log.debug(f"skipping example_name({cur_example_name})")
            continue

        cur_example_dir.mkdir()

        for section, default_filename in [
            ("config", DEFAULT_CONFIG_FILENAME),
            ("workflow", DEFAULT_WORKFLOW_FILENAME),
            ("script", DEFAULT_SCRIPT_FILENAME),
        ]:
            label = f"{pkg_name}.example_{section}_{cur_example_name}"

            log.info(f"{args.example}, label({label})")

            resource = resources.get(f"{section}s", {}).get(label, None)
            if resource is None:
                if section == "script":  # Providing an example script is optional
                    continue

                log.error(f"'No example '{section}' in example({args.example})")
                return errno.EINVAL

            dst = cur_example_dir.joinpath(default_filename)
            if dst.exists():
                log.info(f"skipping dst({dst}); already exists")
                continue

            shutil.copyfile(resource.path, dst)

    return 0


def cli_version(args):
    """Print version and exit"""

    print(f"cijoe {cijoe.core.__version__}")

    return 0


def cli_workflow(args):
    """Process workflow"""

    log.info("cli: run")
    log.info(f"workflow: {args.workflow}")
    log.info(f"configs: {args.config}")
    log.info(f"output: {args.output}")

    cli_archive(args)

    state_path = args.output / "workflow.state"
    if state_path.exists():
        log.error(f"aborting; output({args.output}) directory already exists")
        return errno.EPERM

    os.makedirs(args.output)
    shutil.copyfile(args.workflow, args.output / "workflow.orig")
    if len(args.config) > 1:
        for i, c in enumerate(args.config):
            shutil.copyfile(c, args.output / f"config{i}.orig")

    create_combined_toml(args.config, args.output / "config.orig")
    args.config = args.output / "config.orig"

    config = Config(args.config)
    errors = config.load()
    if errors:
        log_errors(errors)
        log.error(f"failed: Config({args.config}).load()")
        return errno.EINVAL

    workflow = Workflow(args.workflow)

    errors = workflow.load(args, config)
    if errors:
        log_errors(errors)
        log.error("workflow.load(): see errors above or run 'cijoe -i'")
        return errno.EINVAL

    workflow.state["tag"] = args.tag
    step_names = [step["name"] for step in workflow.state["steps"]]
    for step_name in args.step:
        if step_name in step_names:
            continue

        log.error(f"step({step_name}) not in workflow")
        return errno.EINVAL

    resources = get_resources()

    # pre-load scripts and augment state with step-descriptions.
    # TODO: for some reason when mod.__doc__ is None, then the docstring from a previous
    # mod trickles in. This should be fixed...
    for step in workflow.state["steps"]:
        script_ident = step["uses"]
        resources["scripts"][script_ident].load()

        step["description"] = "Undocumented"
        if resources["scripts"][script_ident].mod.__doc__:
            step["description"] = str(resources["scripts"][script_ident].mod.__doc__)

    workflow.state["status"]["started"] = time.time()

    cijoe = Cijoe(config, args.output, args.monitor)
    fail_fast = cijoe.getconf("cijoe.workflow.fail_fast", False)

    for step in workflow.state["steps"]:
        log.info(f"step({step['name']}) - begin")

        begin = time.time()
        step["status"]["started"] = begin
        workflow.state_dump(args.output / Workflow.STATE_FILENAME)

        cijoe.set_output_ident(step["id"])
        os.makedirs(os.path.join(cijoe.output_path, step["id"]), exist_ok=True)

        if args.step and step["name"] not in args.step:
            step["status"]["skipped"] = 1
        else:
            script_ident = step["uses"]
            script = resources["scripts"][script_ident]

            arguments = []
            if "with" in step:
                for k, v in step["with"].items():
                    if type(v) is list:
                        arguments += [f"--{k}", *[f"{el}" for el in v]]
                    else:
                        if type(v) is bool:
                            v = str(v).lower()
                        arguments += [f"--{k}", f"{v}"]

            parser = argparse.ArgumentParser()
            if script.argparser_func:
                script.argparser_func(parser)
            script_args = parser.parse_args(arguments)
            args = parser.parse_args(arguments, namespace=args)

            try:
                err = script.func(args, cijoe)
                if err:
                    log.error(f"script({script_ident}) : err({err})")
                step["status"]["failed" if err else "passed"] = 1
            except KeyboardInterrupt:
                log.exception(f"script({script_ident}) : failed")
                step["status"]["failed"] = 1
            except Exception:
                log.exception(f"script({script_ident}) : failed")
                step["status"]["failed"] = 1
            finally:
                for k in vars(script_args):
                    delattr(args, k)

        for key in ["failed", "passed", "skipped"]:
            workflow.state["status"][key] += step["status"][key]

        step["status"]["elapsed"] = time.time() - begin
        workflow.state["status"]["elapsed"] += step["status"]["elapsed"]

        for text, status in step["status"].items():
            if text != "elapsed" and status:
                log.info(f"step({step['name']}) : {text}")

        if step["status"]["failed"] and fail_fast:
            log.error(f"exiting, fail_fast({fail_fast})")
            break

    workflow.state_dump(args.output / Workflow.STATE_FILENAME)

    err = errno.EIO if workflow.state["status"]["failed"] else 0
    if err:
        log.error("one or more steps failed")

    return err


def create_adhoc_workflow(args):
    target = args.script_name
    if target.endswith(".py"):
        path = Path(target)
        target = path.stem

    resources = get_resources()

    template_path = resources["templates"]["core.example-tmp-workflow.yaml"].path
    jinja_env = jinja2.Environment(
        autoescape=True, loader=jinja2.FileSystemLoader(template_path.parent)
    )
    template = jinja_env.get_template(template_path.name)

    with tempfile.NamedTemporaryFile() as workflow:
        setattr(args, "workflow", Path(workflow.name))
        setattr(args, "step", [])

        content = template.render(steps=[target])
        workflow.write(bytes(content, "utf-8"))
        workflow.seek(0)

        sys.exit(main(args))


def parse_args():
    """Parse command-line interface."""

    parent_dirs = set()
    is_workflow = False

    for i, argv in enumerate(sys.argv):
        if i == 0 or sys.argv[i - 1].startswith("-"):
            continue
        if argv.endswith(".py"):
            path = Path(argv).resolve()
            parent_dirs.add(path.parent)
            sys.argv[i] = path.stem
            break
        if argv.endswith(".yaml"):
            is_workflow = True
            sys.argv.insert(i, "--workflow")
            sys.argv.insert(i, "workflow_path")
            break

    resource_scripts = get_resources(list(parent_dirs))["scripts"]

    parent_parser = argparse.ArgumentParser(add_help=False)

    run_group = parent_parser.add_argument_group(
        "run", "Options for running a workflow script."
    )

    run_group.add_argument(
        "step",
        nargs="*",
        default=[],
        help="Given a workflow, the steps of the workflow it should run. If none are given, all steps are run.",
    )

    run_group.add_argument(
        "--workflow",
        "-w",
        type=Path,
        default=Path(
            os.environ.get("CIJOE_DEFAULT_WORKFLOW", DEFAULT_WORKFLOW_FILENAME)
        ),
        help=argparse.SUPPRESS,
    )
    run_group.add_argument(
        "--config",
        "-c",
        type=Path,
        action=AppendDefaultAction,
        default=[],
        help=f"Path a Configuration file. Multiple can be specified. (default: {os.environ.get('CIJOE_DEFAULT_CONFIG', DEFAULT_CONFIG_FILENAME)})",
    )
    run_group.add_argument(
        "--output",
        "-o",
        type=Path,
        default=default_output_path(),
        help="Path to output directory.",
        action=OverrideDefaultAction,
    )
    run_group.add_argument(
        "--log-level",
        "-l",
        action="append_const",
        const=1,
        help="Increase log-level. Provide '-l' for info and '-ll' for debug.",
    )
    run_group.add_argument(
        "--monitor",
        "-m",
        action="store_true",
        help="Dump command output to stdout",
    )
    run_group.add_argument(
        "--no-report",
        "-n",
        action="store_true",
        help="Skip the producing, and opening, a report at the end of the workflow-run",
    )
    run_group.add_argument(
        "--skip-report",
        "-s",
        action="store_false",
        help="Skip the report opening at the end of the workflow-run",
    )
    run_group.add_argument(
        "--tag",
        "-t",
        type=str,
        action="append",
        help="Tags to identify a workflow-run."
        " This will be prefixed while storing in archive",
    )

    utils_group = parent_parser.add_argument_group(
        "utilities", "Workflow, and workflow-related utilities"
    )
    utils_group.add_argument(
        "--archive",
        "-a",
        action="store_true",
        help="Move the output at '-o / --output' to 'cijoe-archive/YYYY-MM-DD_HH:MM:SS",
    )
    utils_group.add_argument(
        "--produce-report",
        "-p",
        action="append_const",
        const=1,
        help="Produce report, and open it, for output at '-o / --output' and exit.",
    )
    utils_group.add_argument(
        "--integrity-check",
        "-i",
        action="store_true",
        help="Check integrity of workflow given as positional argument and exit.",
    )
    utils_group.add_argument(
        "--resources",
        "-r",
        action="store_true",
        help="List collected resources and exit.",
    )
    utils_group.add_argument(
        "--example",
        "-e",
        type=str,
        nargs="?",
        const="list_examples_and_exit",
        default=None,
        help=(
            "Emits the given example. When no example is given, "
            "then it prints a list of available examples."
        ),
    )
    utils_group.add_argument(
        "--version",
        "-v",
        action="store_true",
        help="Print the version number of 'cijoe' and exit.",
    )

    parser = argparse.ArgumentParser(
        prog=Path(sys.argv[0]).stem,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=[parent_parser],
    )

    subparsers = parser.add_subparsers()

    # Create subparser for workflows
    subparsers.add_parser(
        "workflow_path", parents=[parent_parser], help="Path to a cijoe workflow file."
    )
    subparsers.add_parser("script_path", help="Path to a cijoe script.")
    subparsers.add_parser(
        "script_name",
        help="Name of a cijoe script. You can see all reachable cijoe scripts with command 'cijoe -r'.",
    )

    # Create subparser for scripts
    # Adding the subparser requires loading the whole script, so
    # we only add the subparser to the script given in the arguments
    if not is_workflow:
        for i, argv in enumerate(sys.argv):
            if i == 0 or argv.startswith("-") or sys.argv[i - 1].startswith("-"):
                continue

            # The first positional argument is the script identifier
            ident = argv

            script = resource_scripts.get(ident, None)
            if not script:
                log.error(f"Invalid target({ident})")
                return errno.EINVAL, None
            script.load()
            help_text = (
                next(line for line in script.docs.splitlines() if line)
                if script.docs
                else ""
            )
            subparser = subparsers.add_parser(
                ident,
                parents=[parent_parser],
                help=help_text,
                epilog=script.docs,
                formatter_class=argparse.RawTextHelpFormatter,
            )
            subparser.add_argument(
                "--script-name", default=ident, help=argparse.SUPPRESS
            )
            if script.argparser_func:
                script.argparser_func(subparser)

            break

    args = parser.parse_args()

    levels = [log.ERROR, log.INFO, log.DEBUG]
    log.basicConfig(
        format="%(levelname)s:%(module)s:%(funcName)s(): %(message)s",
        level=levels[
            min(sum(args.log_level), len(levels) - 1) if args.log_level else 0
        ],
        force=True,
    )

    return 0, args


def main(args=None):
    """Main entry point for the CLI"""

    if args is None:
        err, args = parse_args()
        if err:
            log.error("Couldn't parse args; exiting")
            return err
        if getattr(args, "script_name", None):
            # Running stand-alone script
            create_adhoc_workflow(args)

    if args.resources:
        return cli_resources(args)

    if args.example:
        return cli_example(args)

    if args.version:
        return cli_version(args)

    if args.produce_report:
        return cli_produce_report(args)

    if args.archive:
        return cli_archive(args)

    if not getattr(args, "workflow", None):
        log.error("No target given; exiting")
        return errno.EINVAL

    if not args.config:
        args.config = [
            Path(os.environ.get("CIJOE_DEFAULT_CONFIG", DEFAULT_CONFIG_FILENAME))
        ]

    path = search_for_file(args.workflow)
    if path is None:
        log.error(f"workflow({args.workflow}) does not exist; exiting")
        return errno.EINVAL
    args.workflow = path

    tmp = args.config
    args.config = []
    for argv in tmp:
        path = search_for_file(argv)
        if path is None:
            log.error(f"config({argv}) does not exist; exiting")
            return errno.EINVAL

        args.config.append(path)

    if args.integrity_check:
        return cli_integrity_check(args)

    # At this point all Path objects are resolved, except for output, do that
    # that here.
    args.output = args.output.resolve()

    err = cli_workflow(args)

    if not args.no_report:
        report_err = cli_produce_report(args)
        err = err if err else report_err

    return err



"""
    null_blk module, helpers to load/unload null block instances

    To use it, one must have permissions to modprobe, rmmod and modify the /syscfg

    NOTE: This is **fully** re-targetable, that is, changing transport changes where it
    is running.

    For reference, see: https://docs.kernel.org/block/null_blk.html

    TODO
    - Implement initialization of nullblk instances via /sys/kernel, this allows for
      different device instances instead of N instances with the same configuration.
      For example, one can instantiate a regular block-device as well as a zoned
      block-device.

    retargetable: True
"""

NULLBLK_MODULE_NAME = "null_blk"
NULLBLK_SYSPATH = "/sys/kernel/config/nullb"


def insert(cijoe, config=None):
    """Load the 'null_blk' kernel module using parameters defined in the config"""

    if config is None:
        config = cijoe.config.options.get("null_blk")

    nullblk_params = (
        " ".join([f"{k}={v}" for k, v in config.items()])
        if config.get("nr_devices")
        else ""
    )

    return cijoe.run(f"modprobe {NULLBLK_MODULE_NAME} {nullblk_params}")


def remove(cijoe):
    """Remove the null_blk kernel module"""

    # This can be used when instanttation via SYSPATH, however, commented out as it is
    # not useful yet.
    # cijoe.run(f"rmdir {NULLBLK_SYSPATH}/nullb*")

    return cijoe.run(f"modprobe -r {NULLBLK_MODULE_NAME}")



import errno
import gzip
import hashlib
import logging as log
import lzma
import shutil
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import Optional

import requests

ENCODING = "UTF-8"


def sanitize_ident(ident: str):
    """Naive ident sanitizer"""

    for illegal in r'/\|!?[]<>.,:;"*':
        ident = ident.replace(illegal, "_")

    return ident


def download(url: str, path: Path):
    """Downloads a file over http(s), returns (err, path)."""

    path = Path(path).resolve()
    if path.is_dir():
        path = path / url.split("/")[-1]
    if not (path.parent.is_dir() and path.parent.exists()):
        return errno.EINVAL, path

    with requests.get(url, stream=True) as request:
        request.raise_for_status()
        with path.open("wb") as local:
            for chunk in request.iter_content(chunk_size=8192):
                local.write(chunk)

    return 0, path


def get_checksums_from_url(url_checksum: str):
    """
    Downloads checksum(s) from given url to a temporary directory, returns
    the hashing algorithm and the contents of the file. The algorithm is
    found based on the filename of the downloaded checksum file.
    Returns (err, checksums, algorithm).
    """

    dir = TemporaryDirectory()
    err, path_checksum = download(url_checksum, Path(dir.name).resolve())
    if err:
        log.error(f"download({url_checksum}), {path_checksum}: failed")
        return err, None, None

    with open(path_checksum, "r") as checksum_f:
        # Loop through valid hash algorithms to find the one that matches the
        # filename of the checksum
        for algorithm in hashlib.algorithms_guaranteed:
            if algorithm in path_checksum.name.lower():
                return 0, checksum_f.read(), algorithm

    log.error(
        "error: downloaded checksum does contain the name of a valid hash algorithm"
    )
    return 1, None, None


def download_and_verify(url: str, url_checksum: str, path: Path):
    """
    Downloads a file over http(s). The file is only downloaded if checksums are
    not equal and if the checksum at url_checksum matches the downloaded file,
    returns (err, path).
    """

    path = Path(path).resolve()
    if path.is_dir():
        path = path / url.split("/")[-1]
    if not (path.parent.is_dir() and path.parent.exists()):
        return errno.EINVAL, path

    err, verification, algorithm = get_checksums_from_url(url_checksum)
    if err:
        log.error(f"error when downloading checksum ({url_checksum})")

    # The checksum of the existing file at path
    checksum = None
    checksum_path = path.with_suffix(path.suffix + f".{algorithm}sum")
    if path.exists() and checksum_path.exists():
        with open(checksum_path, "r") as f:
            checksum = f.read()

    # If the file does not already exists or if checksums do not match,
    # download from url
    if not checksum or checksum not in verification:
        log.info(f"Downloading file from {url}")

        with NamedTemporaryFile() as dwnld:
            err, dwnld_path = download(url, Path(dwnld.name).resolve())
            if err:
                log.error(f"download({url}), {dwnld_path}: failed")
                return err, None

            # Check that downloaded checksum file contains the checksum of
            # the downloaded file
            new_checksum = hashlib.file_digest(dwnld, algorithm).hexdigest()
            if new_checksum not in verification:
                log.error(
                    f"error while downloading file: checksum ({new_checksum}) not in checksum file:\n{verification}"
                )
                return 1, None

            # Move contents of temporary file to the given path
            shutil.copyfile(dwnld_path, path)
            with open(checksum_path, "w") as f:
                f.write(new_checksum)

    return 0, path


def decompress_file(src: Path, dst: Optional[Path] = None):
    """
    Decompress the file at the given src path. If no destination path is given,
    the file is decompressed at the src path without the compression suffix.
    Returns (err, decompressed_path). Throws an error if the file ending does
    not match any of the supported compression types.

    Supported compression types: .xz, .gz
    """

    supported_compression_types = [".xz", ".gz"]

    if src.suffix not in supported_compression_types:
        log.error(f"Unsupported compression suffix: {src.suffix}")
        return 1, None

    src = src.resolve()

    if not dst:
        dst = Path(str(src).replace(src.suffix, "")).resolve()

    if dst.exists():
        log.info(f"File({dst}) already exists, skipping decompression")
        return 0, dst

    decompressor = lzma if src.suffix == ".xz" else gzip

    with decompressor.open(src, mode="rb") as fsrc:
        with open(dst, "xb") as fdst:
            shutil.copyfileobj(fsrc, fdst)
            log.info(f"Decompressed file({src}) to destination({dst})")

    return 0, dst



"""
cmdrunner
=========

Executes a list of commands in the given order. Note that multi-line commands are not
support, each line or list of strings are treated as individual commands.

Retargetable: True
------------------
"""

import errno
import logging as log
from argparse import ArgumentParser


def add_args(parser: ArgumentParser):
    parser.add_argument(
        "--commands", nargs="+", type=str, help="The commands to be run"
    )
    parser.add_argument(
        "--transport",
        type=str,
        default=None,
        help=(
            "The key of the transport from the cijoe config file on which the commands should be run. "
            "Use 'initiator' if the commands should be run locally. "
            "Defaults to the first transport in the config file ('initiator' if none are defined)."
        ),
    )


def main(args, cijoe):
    """Run commands one at a time via cijoe.run()"""

    err = 0
    if "commands" not in args:
        log.error("missing step-argument: with.commands")
        return errno.EINVAL

    for cmd in args.commands:
        err, state = cijoe.run(cmd, transport_name=args.transport)
        if err:
            break

    return err



#!/usr/bin/env python3
"""
Start a qemu guest
==================

Starts the qemu guest with the given guest name. Fails if the guest is not up
within 180 seconds.

Retargetable: False
-------------------
"""
import errno
import logging as log
from argparse import ArgumentParser

from cijoe.qemu.wrapper import Guest


def add_args(parser: ArgumentParser):
    parser.add_argument("--guest_name", type=str, help="Name of the qemu guest.")


def main(args, cijoe):
    """Start a qemu guest"""

    if "guest_name" not in args:
        log.error("missing argument: guest_name")
        return 1

    guest = Guest(cijoe, cijoe.config, args.guest_name)

    err = guest.start()
    if err:
        log.error(f"guest.start() : err({err})")
        return err

    started = guest.is_up(timeout=180)
    if not started:
        log.error("guest.is_up() : False")
        return errno.EAGAIN

    return 0



"""
cijoe example script
====================

The script is a modified "Hello, World!" example. It repeatedly prints a message a set
number of times and allows parameterization of the message content.

The purpose of this script is to demonstrate how to run commands and supply input to the
script using a configuration file, environment variables, command-line arguments and 
workflow step arguments.

An example of using the core infrastructure of cijoe:

* cijoe.run(command)

  - Error-handling; checking return-code
  - Output processing state.output()

Input is given to scripts via configuration-files, environment variables and from
workflow-step-arguments, this is demonstrated as the first thing in the script.

cijoe also has primitives for transferring data:

* cijoe.get(src, dst)

  - Transfer 'src' directory or file from **target** to 'dst' on **initiator** 

* cijoe.put(src, dst)

  - Transfer 'src' directory or file from **initiator** to 'dst' on **target** 

These are not used in the example code below, but you can experiment and try adding
them yourself.
"""

import logging as log
from argparse import ArgumentParser, Namespace

from cijoe.core.command import Cijoe


def add_args(parser: ArgumentParser):
    """Optional function for defining command-line arguments for this script"""
    parser.add_argument(
        "--repeat",
        type=int,
        default=1,
        help="Amount of times the message will be repeated",
    )


def main(args: Namespace, cijoe: Cijoe):
    """Entry-point of the cijoe-script"""

    # Grab message from the configuration-file
    message = cijoe.getconf("example.message", "Hello World!")

    # When executed via workflow, grab the step-argument
    repeat = args.repeat
    if repeat < 1:
        log.error(f"Invalid step-argument: repeat({repeat}) < 1")
        return 1

    log.info(f"Will echo the message({message}), repeat({repeat}) times")

    # Now, execute a command that echoes the 'message' 'repeat' number of times
    for _ in range(1, repeat + 1):
        err, state = cijoe.run(f"echo '{message}'")
        if "Hello" not in state.output():
            log.error("Something went wrong")
            return 1

    log.info("Success!")

    return 0



#!/usr/bin/env python3
"""
Create docker image using a disk image
======================================

This will produce dockers images for all the system images described in config. section:

* ``system_imaging.images``

You can reduce this by providing a case-incensitive fnmatch pattern as input to the
script via workflow, such as these::

    # This will build all images
    with:
      pattern: "*"

    # This will build all those starting with debian
    with:
      pattern: "debian*"

Retargetable: False
-------------------

This script only runs on the iniator; due to the use of 'shutil', 'download' etc.
"""
import logging as log
import shutil
from argparse import ArgumentParser
from fnmatch import fnmatch
from pathlib import Path

from cijoe.core.misc import download
from cijoe.core.resources import get_resources


def add_args(parser: ArgumentParser):
    parser.add_argument("--pattern", type=str, help="Pattern for image names to build")


def dockerimage_from_diskimage(cijoe, image):
    resources = get_resources()

    err, state = cijoe.run_local("mktemp -d")  # Create temporary directory
    if err:
        log.error("Failed creating workdir")
        return

    workdir = Path(state.output().strip())
    mountpoint = workdir / "mount"

    shutil.copyfile(
        str(resources.get("auxiliary", {}).get("system_imaging.Dockerfile").path),
        str(workdir / "Dockerfile"),
    )
    shutil.copyfile(
        str(resources.get("auxiliary", {}).get("system_imaging.dockerignore").path),
        str(workdir / ".dockerignore"),
    )

    err, _ = cijoe.run_local(f'mkdir -p "{mountpoint}"')

    commands = [
        f"guestmount -a {image['disk']['path']} -i --ro {mountpoint}",
        f"docker build -t {image['docker']['name']}:{image['docker']['tag']} -f {workdir}/Dockerfile {mountpoint}",
        f"guestunmount {mountpoint}",
    ]
    for command in commands:
        err, _ = cijoe.run_local(command)
        if err:
            log.error(f"command({command}); err({err})")
            return err

    cijoe.run_local("docker builder prune -a -f")
    cijoe.run_local(f'echo "Needs cleanup!" && find {workdir}')
    cijoe.run_local(
        f'echo "Run with: docker run -it {image["docker"]["name"]}:{image["docker"]["tag"]} bash"'
    )

    return 0


def main(args, cijoe):
    """Create a docker image using the content of a .qcow2 image"""

    if "pattern" not in args:
        log.error("missing step-argument: with.pattern")
        return 1
    pattern = args.pattern

    log.info(f"Got pattern({pattern})")

    entry_name = "system-imaging.images"
    images = cijoe.getconf(entry_name, {})
    if not images:
        log.error(f"missing: '{entry_name}' in configuration file")
        return 1

    count = 0
    for image_name, image in cijoe.getconf("system-imaging.images", {}).items():
        if not fnmatch(image_name.lower(), pattern.lower()):
            log.info(f"image_name({image_name}); did not match pattern({pattern}")
            continue

        log.info(f"image_name({image_name}); matched pattern({pattern})")

        err = dockerimage_from_diskimage(cijoe, image)
        if err:
            log.error(f"failed dockerimage_from_diskimage(); err({err})")
            return err

        count += 1

    if not count:
        log.error(f"did not build anything, count({count}); invalid with.pattern?")
        return err

    return 0



#!/usr/bin/env python3
"""
Install qemu
============

This script installs qemu on the host machine. It is assumed that qemu has been
built in advance at the given repository path.

Configuration
-------------

* ``qemu.repository.path``: str

  Path to the qemu repository on the target machine.

Retargetable: False
-------------------
"""
import errno
import logging as log
from pathlib import Path


def main(args, cijoe):
    """Install qemu"""

    path = cijoe.getconf("qemu.repository.path", None)
    if not path:
        log.error("missing 'qemu.repository.path' in configuration file")
        return errno.EINVAL

    build_dir = Path(path) / "build"

    err, _ = cijoe.run_local("make install", cwd=build_dir)

    return err



"""
Linux Custom Kernel as Debian Package
=====================================

There are a myriad of ways to build and install a custom Linux kernel. This worklet
builds it as a Debian package. The generated .deb packages are stored in
cijoe.output_path.

Retargetable: False
-------------------

It is intended to be run "locally" since, currently the collection of the generated
.debs are not retrieved via cijoe.get(), doing so would make it retargetable.
"""

import logging as log
from argparse import ArgumentParser, _StoreAction
from pathlib import Path


def add_args(parser: ArgumentParser):
    class StringToBoolAction(_StoreAction):
        def __call__(self, parser, namespace, values, option_string=None):
            setattr(namespace, self.dest, values == "true")

    parser.add_argument(
        "--local_version",
        type=str,
        default="custom",
        help="Path to local version of kdebs",
    )
    parser.add_argument(
        "--run_local",
        choices=["true", "false"],
        default=True,
        action=StringToBoolAction,
        help="Whether or not to execute in the same environment as 'cijoe'.",
    )


def main(args, cijoe):
    """Configure, build and collect the build-artifacts"""

    path = cijoe.getconf("linux.repository.path")
    if not path:
        log.error("missing config: linux.repository.path")
        return 1

    repos = Path(path).resolve()
    err, _ = cijoe.run(f"[ -d {repos} ]")
    if err:
        return err

    run = cijoe.run_local if args.run_local else cijoe.run

    commands = [
        "[ -f .config ] && rm .config || true",
        "yes " " | make olddefconfig",
        "./scripts/config --disable CONFIG_DEBUG_INFO",
        "./scripts/config --disable SYSTEM_TRUSTED_KEYS",
        "./scripts/config --disable SYSTEM_REVOCATION_KEYS",
        f"yes '' | make -j$(nproc) bindeb-pkg LOCALVERSION={args.local_version}",
        f"mkdir -p {cijoe.output_path}/artifacts/linux",
        f"mv ../*.deb {cijoe.output_path}/artifacts/linux",
        f"mv ../*.changes {cijoe.output_path}/artifacts/linux",
        f"mv ../*.buildinfo {cijoe.output_path}/artifacts/linux",
    ]
    for cmd in commands:
        err, _ = run(cmd, cwd=str(repos))
        if err:
            return err

    return 0



"""
    Wraps qemu binaries: system, qemu-img and provides "guest-control"

    NOTE: this wrapper is "local-only". That is, changing transport does not retarget
    the functionality provided here. Most of the code is utilizing Python modules such
    as shutil, pathlib, psutil, download/requests. To make this re-targetable these
    things must be done via command-line utilities. It is certainly doable, however,
    currently not a priority as the intent is to utilize qemu to produce a virtual
    machine to serve as a 'target' for tests.
"""

import errno
import logging as log
import os
import shutil
import time
from pathlib import Path
from pprint import pformat
from typing import Optional

import psutil

from cijoe.core.misc import download_and_verify


def qemu_img(cijoe, args=""):
    """Helper function wrapping around 'qemu-img'"""

    qemu_img_bin_default = "qemu-img"

    qemu_img_bin = cijoe.getconf("qemu.img_bin", None)
    if qemu_img_bin is None:
        log.error("Could not determine 'qemu-img' binary")
        qemu_img_bin = qemu_img_bin_default

    log.info(
        f"qemu_img_bin({qemu_img_bin}), qemu_img_bin_default({qemu_img_bin_default})"
    )

    return cijoe.run_local(f"{qemu_img_bin} {args}")


def qemu_system(cijoe, system_label, args=""):
    """Resolves and invokes the qemu-system by its system_label e.g. 'x86_64'"""

    system_bin = cijoe.getconf(f"qemu.systems.{system_label}.bin", None)
    if system_bin is None:
        log.error(f"Cannot determine system using system_label({system_label})")
        return errno.EINVAL, None

    return cijoe.run_local(f"{system_bin} {args}")


class Guest(object):
    def __init__(self, cijoe, config, guest_name):
        """."""

        qemu_config = config.options.get("qemu", {})
        qemu_guests = qemu_config.get("guests", [])
        if not (qemu_config and qemu_guests):
            raise ValueError(f"Invalid qemu_config({pformat(qemu_config)})")

        guest_config = qemu_guests.get(guest_name, None)
        if not guest_config:
            raise ValueError(f"Invalid guest_config({pformat(guest_config)})")

        guest_path = guest_config.get("path", None)
        if not (guest_config and guest_path):
            log.error(f"invalid qemu_config({pformat(qemu_config)})")
            raise ValueError("Invalid configuration")

        self.cijoe = cijoe
        self.qemu_config = qemu_config
        self.guest_config = guest_config
        self.guest_path = Path(guest_path).resolve()

        self.bios_img = self.guest_path / "bios.img"
        self.boot_img = self.guest_path / "boot.img"
        self.pid = self.guest_path / "guest.pid"
        self.monitor = self.guest_path / "monitor.sock"
        self.serial = self.guest_path / "serial.output"

    def image_create(self, filename, fmt="raw", size="8GB"):
        """
        Creates an image-file in the guest_path. Returns 0 on succes, errno to
        indicate the error.
        """

        img_path = self.guest_path / filename
        err, _ = qemu_img(self.cijoe, f"create -f {fmt} {img_path} {size}")

        return err

    def is_initialized(self):
        """Check that the guest is initialized"""

        return self.guest_path.exists()

    def is_running(self):
        """Check whether the guest is running"""

        pid = self.get_pid()

        return pid and psutil.pid_exists(pid)

    def get_pid(self):
        """Returns pid from 'guest.pid', returns 0 when 'guest.pid' is not found"""

        if not self.pid.exists():
            return 0

        with self.pid.open() as pidfile:
            pid = int(pidfile.read().strip())

        return pid

    def initialize(self, diskimage_path: Optional[Path] = None):
        """Create a 'home' for the guest'"""

        os.makedirs(self.guest_path, exist_ok=True)

        if diskimage_path:
            err, _ = self.cijoe.run_local(f"cp {diskimage_path} {self.boot_img}")
            if err:
                log.error(f"Failed copying diskimage({diskimage_path}); err({err})")
                return err

        if self.guest_config.get("system_label", None) != "aarch64":
            return 0

        # aarch64
        bios_path = None
        for location in [
            "/usr/share/edk2/aarch64/QEMU_EFI.fd",
            "/usr/share/qemu-efi-aarch64/QEMU_EFI.fd",
        ]:
            if (cand_path := Path(location)).exists():
                bios_path = cand_path

        if bios_path is None:
            log.error("Could not find aarch64 bios/firmware")
            return errno.EINVAL

        err, _ = self.cijoe.run_local(f"cp {bios_path} {self.bios_img}")
        if err:
            log.error("copy failed")
            return errno.EINVAL

        err, _ = self.cijoe.run_local(f"truncate -s 64m {self.bios_img}")
        if err:
            log.error("truncate failed")
            return errno.EINVAL

        return 0

    def is_up(self, timeout=120):
        """Wait at most 'timeout' seconds for the guest to print 'login' to serial"""

        if not self.is_running():
            log.error("is_running(), check via pid, is False")
            return False

        began = time.time()
        while True:
            enter = time.time()
            try:
                if "login:" in self.serial.read_text(
                    encoding="utf-8", errors="replace"
                ):
                    return True
            except Exception as exc:
                log.error(f"{exc}")

            now = time.time()
            elapsed_iter = now - enter
            elapsed_total = now - began

            if elapsed_iter < 5.0:
                time.sleep(5.0 - elapsed_iter)
            if elapsed_total > timeout:
                log.error(f"System did not come up within timeout({timeout}) seconds")
                return False

    def start(self, daemonize=True, extra_args=[]):
        """."""

        system_label = self.guest_config.get("system_label", None)
        if system_label is None:
            log.error(f"invalid qemu.guests.GUEST.system_label({system_label})")
            return errno.EINVAL

        args = []

        system_args = self.guest_config.get("system_args", {})

        # Key-word arguments: config.qemu.guest.system_args.kwa
        for key, value in system_args.get("kwa", {}).items():
            args.append(f"-{key}")
            if isinstance(value, dict):
                args.append(next((f"{opt}={val}" for opt, val in value.items())))
            else:
                args.append(str(value))

        #
        # Managed: when 'bios.img' exists, add it as a 'pflash'
        #
        if self.bios_img.exists():
            args += [
                "-drive",
                f"file={self.bios_img},format=raw,if=pflash,readonly=on",
            ]

        #
        # Managed: when 'boot.img' exists, add it is as boot-drive
        #
        if self.boot_img.exists():
            args += [
                "-blockdev",
                f"qcow2,node-name=boot,file.driver=file,file.filename={self.boot_img}",
            ]
            args += ["-device", "virtio-blk-pci,drive=boot"]

        # Process Management stuff
        args += ["-pidfile", str(self.pid)]
        args += ["-monitor", f"unix:{self.monitor},server,nowait"]

        if daemonize:
            args += ["-display", "none"]
            args += ["-serial", f"file:{self.serial}"]
            args += ["-daemonize"]
        else:
            args += ["-nographic"]
            args += ["-serial", "mon:stdio"]

        #
        # Managed: config.qemu.guest.system_args.host_share
        #
        host_share = system_args.get("host_share", None)
        if host_share:
            host_share = Path(host_share).resolve()
            args += [
                "-virtfs",
                "fsdriver=local,id=fsdev0,security_model=mapped,mount_tag=hostshare"
                f",path={host_share}",
            ]

        #
        # Managed: config.qemu.guest.system_args.tcp_forward
        #
        tcp_forward = system_args.get("tcp_forward", None)
        if tcp_forward:
            host_addr = tcp_forward.get("host_address", "127.0.0.1")
            host_port = tcp_forward.get("host", None)
            guest_port = tcp_forward.get("guest", None)

            if None in [host_port, guest_port]:
                log.error(f"Invalid tcp_forward({tcp_forward})")
                return errno.EINVAL

            args += [
                "-netdev",
                f"user,id=n1,ipv6=off,hostfwd=tcp:{host_addr}:{host_port}-:{guest_port}",
            ]
            args += ["-device", "virtio-net-pci,netdev=n1"]

        #
        # Arguments provided via wrapper-class
        #
        args += extra_args

        #
        # Raw: config.qemu.guest.system_args.raw
        #
        args += [system_args.get("raw", "")]

        err, _ = qemu_system(self.cijoe, system_label, " ".join(args))
        if err:
            log.error(f"qemu_system failed with err({err})")
            return err

        return 0

    def wait_for_termination(self, timeout=60):
        """
        Wait for 'timeout' seconds for the qemu guest to terminate. This method does
        not itself terminate the qemu guest. Returns (err: int, terminated: bool) where
        'terminated' defines whether the qemu guest has been terminated or not.
        """
        err = 0

        start = time.time()
        while True:
            if not self.get_pid():
                return 0, True
            if time.time() - start > timeout:
                break
            time.sleep(1)

        return err, False

    def kill(self):
        """Shutdown qemu guests by killing the process using the 'guest.pid'"""

        err = 0

        try:
            pid = self.get_pid()
            if pid:
                qemu_proc = psutil.Process(pid)
                qemu_proc.terminate()

                err, terminated = self.wait_for_termination(60)
                if err:
                    log.error(f"Could not terminate process({pid})")
                    return err

                if not terminated:
                    qemu_proc.kill()
        except psutil.NoSuchProcess:
            log.info("Got 'NoSuchProcess', that is OK, continue.")

        return err
